<?php
define('SUCCESS_STRING', 'setupscript_success');

$_SC_LOG_FILE = NULL;
function _Log ($msg)
{
    global $_SC_LOG_FILE;

    if( $_SC_LOG_FILE == NULL )
    {
        $SV = SetupVars::get();
        $_SC_LOG_FILE = $SV->INSTALL_DIR.'/piscript.log';
    }
    
    //    return;
    $args = func_get_args();
    array_shift($args);
    $str = date('[H:i:s]') . ' ' . @vsprintf($msg, $args) . "\n";

    file_put_contents( $_SC_LOG_FILE, $str, FILE_APPEND );
    //echo $str;
}


class SetupVars
{
    public $SITE_NAME;
    public $ADM_USERNAME;
    public $ADM_PASSWORD;
    public $ADM_REALNAME;
    public $ADM_EMAIL;
    public $DB_NAME;
    public $DB_USERNAME;
    public $DB_PASSWORD;
    public $APACHE_PORT;
    public $MYSQL_PORT;
    public $SITE_DIR;
    public $INSTALL_DIR;
    public $PLATFORM_NAME;
    public $AD_VER;

    public function LoadSetupVars ($file)
    {
        $v = Util::ParseIniFile($file);
        foreach ($v['vars'] as $name => $val)
        {
            $this->$name = $val;
        }
    }

    /**
     * @return SetupVars
     */
    public static function get ()
    {
        if (self::$_inst == null)
        {
            self::$_inst = new SetupVars();
        }
        return self::$_inst;
    }

    private function __construct ()
    {
    }
    private static $_inst;
}


class Setup
{
    // some OSX consts
    const ACQUIA_GROUP = 'acquia_drupal';
    const APACHE_USER = 'ad_apache';
    const MYSQL_USER = 'ad_mysql';        
    const APACHE_PLIST = 'com.acquia.drupal.apache';    
    const MYSQL_PLIST = 'com.acquia.drupal.mysql';
    const PLIST_DIR = '/Library/LaunchDaemons';
    

    public  static function Init()
    {
        $SV = SetupVars::get();
        $SV->LoadSetupVars(dirname(__FILE__) . '/setupvars');        
    }
    
    public static function Run ()
    {
        self::Init();
        $SV = SetupVars::get();

        _Log('Post install script started');
        
        //  See AN-5421. User is no longer asked for database settings, so we hardcode them here.
        $SV->DB_NAME = 'acquia_drupal';
        $SV->DB_USERNAME = 'drupaluser';
        $SV->DB_PASSWORD = '';//empty password

        
        self::SubstituteFile($SV->INSTALL_DIR . '/MasterLicense.txt');
        
        self::SubstituteFile($SV->INSTALL_DIR . '/apache/conf/httpd.conf');
        file_put_contents($SV->INSTALL_DIR . '/apache/conf/vhosts.conf', "");
        self::SubstituteFile($SV->INSTALL_DIR . '/mysql/my.cnf');                
        
        self::SubstituteFile($SV->SITE_DIR . '/phpmyadmin/config.inc.php');
        self::SubstituteFile($SV->SITE_DIR . '/phpmyadmin/config.inc.php', 
                array('@@BLOWFISH_SECRET@@' => md5( microtime() ) ) );
        
        $drpSettingsPath = $SV->SITE_DIR . '/acquia-drupal/sites/default';
        
        if ( !file_exists($drpSettingsPath . '/settings.php') && 
              file_exists($drpSettingsPath . '/default.settings.php')  )
        {            
            copy( $drpSettingsPath . '/default.settings.php', $drpSettingsPath . '/settings.php' );
        }
                
        @mkdir($SV->SITE_DIR . '/acquia-drupal/sites/default/files');

        //create phpinfo file
        file_put_contents($SV->SITE_DIR.'/acquia-drupal/phpinfo.php', $GLOBALS['DATA_PHPINFO_PHP_CODE'] );        
        
        if ($SV->PLATFORM_NAME == 'windows')
        {
            self::SubstituteFile($SV->INSTALL_DIR . '/AcquiaDrupalControlPanel/static.ini');
            self::SubstituteFile($SV->INSTALL_DIR . '/AcquiaDrupalControlPanel/dynamic.ini');
            //self::SubstituteFile($SV->INSTALL_DIR . '/AcquiaDrupalControlPanel/svn/svn.cmd');
            
            self::SubstituteFile($SV->INSTALL_DIR . '/php/php.ini', 
                array( '@@TEMP_DIR@@' => rtrim( sys_get_temp_dir(), '\\' ) ) );
            
            self::SubstituteFile($SV->INSTALL_DIR . '/common/start.cmd');
            self::SubstituteFile($SV->INSTALL_DIR . '/common/stop.cmd');
            
            self::SubstituteFile($SV->INSTALL_DIR . '/mysql/bin/mysql.cmd');
            
            //windows mysql doesn't like \'s in ini paths.
            self::SubstituteFile($SV->INSTALL_DIR . '/mysql/my.cnf', array('\\' => '/'));
            
            
            //copy( $SV->INSTALL_DIR.'/php/libmysql.dll', $SV->INSTALL_DIR.'/apache/bin/libmysql.dll' );
            self::SetupWindowsServices();
        }
        else if ($SV->PLATFORM_NAME = 'osx')
        { 
            self::SubstituteFile($SV->INSTALL_DIR . '/apache/bin/envvars');
            self::SubstituteFile($SV->INSTALL_DIR . '/mysql/bin/mysqladmin');
            self::SubstituteFile($SV->INSTALL_DIR . '/mysql/bin/mysqld');
            self::SubstituteFile($SV->INSTALL_DIR . '/mysql/bin/mysql');            
            self::SubstituteFile($SV->INSTALL_DIR . '/mysql/bin/mysqldump');                        
            self::SubstituteFile($SV->INSTALL_DIR . '/php/bin/php.ini');            
            self::SubstituteFile($SV->INSTALL_DIR . '/php/bin/php.ini', 
                array('@@TEMP_DIR@@' => '/tmp' ) );
            
            self::SubstituteFile($SV->INSTALL_DIR . '/common/envvars');
            self::SubstituteFile($SV->INSTALL_DIR . '/common/start');
            self::SubstituteFile($SV->INSTALL_DIR . '/common/stop');
            //chmod ( $SV->INSTALL_DIR . '/common/start', 0755 );
            //chmod ( $SV->INSTALL_DIR . '/common/stop', 0755 );
            self::SetupUnixMysql();
            self::SetupUnixApache();
            $instId = self::SetupOSXServices( $SV->INSTALL_DIR, $SV->SITE_DIR );
            
            self::SubstituteFile($SV->INSTALL_DIR . '/Acquia Drupal Control Panel.app/Contents/MacOS/static.ini');
            self::SubstituteFile($SV->INSTALL_DIR . '/Acquia Drupal Control Panel.app/Contents/MacOS/dynamic.ini');
            $iid = array('@@OSX_INST_ID@@' => $instId);
            self::SubstituteFile($SV->INSTALL_DIR . '/Acquia Drupal Control Panel.app/Contents/MacOS/static.ini', $iid);
            
            // make them available for apache (run on behalf of 'daemon' user)
            chmod($SV->SITE_DIR . '/acquia-drupal/sites/default/settings.php', 0666 );
            chmod($SV->SITE_DIR . '/acquia-drupal/sites/default/files', 0777 );   

        }
        else
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unsupported platform:' . $SV->PLATFORM_NAME);
        }

        
        try
        {
            if( $SV->PLATFORM_NAME != 'osx' )
            {
                self::StartMySQL();
                self::StartApache();                
            }
            else 
            {
                self::WaitMySQLStart();
                self::WaitApacheStart();
            }
            self::CreateDB();
            
            if( $SV->PLATFORM_NAME == 'windows' )
                self::SetupXMail();
            
            self::InstallDrupal();
            
            //self::SetDrupalBaseUrl( $drpSettingsPath . '/settings.php' );
        }
        catch (Exception $e)
        {
            try
            {
                if( $SV->PLATFORM_NAME != 'osx' )
                {
                    self::StopMySQL();
                    self::StopApache();                    
                }
            }
            catch (Exception $__e)
            {
            } // we don't care about exceptions here
            throw $e;
        }

        if( $SV->PLATFORM_NAME != 'osx' )
        {
            self::StopMySQL();
            self::StopApache();           
        }
        else
        {
            $stdout = $stderr =  null;
            //correct file permissions for install dir
            _Log("correct file permissions for install dir");
            $cmdLine = "chmod -R a+r ".escapeshellarg($SV->INSTALL_DIR);
            Util::ExecC($cmdLine, $stdout);
            $cmdLine = "chmod -R g+w ".escapeshellarg($SV->INSTALL_DIR);
            Util::ExecC($cmdLine, $stdout);
        
            //correct file permissions for Site install dir
            _Log("correct file permissions for Site install dir"); 
            $userName = getenv('SUDO_USER')? getenv('SUDO_USER') : getenv('USER');
            if( $userName && $userName != 'root' )
            {
                if( OSXUtil::OSUserExists($userName) && 
                    ( $groupName = OSXUtil::GetOSGroupNameByID( OSXUtil::GetOSUserGroupID( $userName ) ) ) != null )
                {                                        
                    $fullName = $groupName? "$userName:$groupName": $userName;
                    Util::Exec("chown -R $fullName ".escapeshellarg($SV->SITE_DIR), $stdout, $stderr );                 
                }
                else 
                {
                    Util::Exec("chown -R $userName ".escapeshellarg($SV->SITE_DIR), $stdout, $stderr );                 
                }
            }
            chmod($SV->SITE_DIR . '/acquia-drupal/sites/default/settings.php', 0444 );
                        
            _Log("Prepare PHP extension helper scripts");
            self::SetupUnixPhpize();                    
        }
        
        _Log('Post install script completed successfully');
    }
    
    private static function SetDrupalBaseUrl( $settingsFile, $host = null, $port = null )
    {
        $sv = SetupVars::get();
        if( $host === null )
            $host = 'localhost';
        if( $port === null )
            $port = $sv->APACHE_PORT;
        
        Util::RegexReplaceInFile(
            $settingsFile, 
            '/^[^*]*\$base_url.*$/m',
            sprintf('$base_url=\'http://%s:%d\';', $host, $port)  );
   
    }

    public static function FinishSiteInstall( $args )
    {
        //print_r($args); return;
        self::Init();
        _Log('Finish site install');
        
        DrupalCli::Setup($_SERVER);
        $installer = new DrupalInstaller();
    
        $installer->profile = base64_decode( @$args['profile'] );
        $installer->site_name = base64_decode( @$args['site_name'] );
        $installer->username = base64_decode( @$args['username'] );
        $installer->password = base64_decode( @$args['password'] );
        $installer->email = base64_decode( @$args['email'] );

        $installer->install2( base64_decode( @$args['baseurl']) );
    }
    
    /**
     * 
     * @return  array(apachename, mysqlname);
     */
    private static function SetupWindowsServices ()
    {
        _Log('Setting up windows services');
        
        $APACHE_SCM_ACCESS_DENIED = 720005; //TODO: found out experimentally for 2.2. May have problems with other versions 
        $TRY_NUM = 20;
        
        //setup apache
        $dir = realpath(SetupVars::get()->INSTALL_DIR . '/apache/bin');
        $apacheSvcName = '';
        for ($i = 0; $i < $TRY_NUM; $i ++)
        {
            $apacheSvcName = "ad_apache$i";
            $stdout = $stderr = '';
            $res = Util::Exec(".\\httpd.exe -k install -n $apacheSvcName", $stdout, $stderr, $dir);
            if ($res == $APACHE_SCM_ACCESS_DENIED || stripos($stderr, "Access is denied") !== false)
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Access to SCM is denied');
            }
            else if ($res != 0 && stripos($stderr, "successfully installed") !== false)
            {
                // this may happen if there is an error in httpd.conf for example. (found out experimentally)
                $out = $err = '';
                $res = Util::Exec(".\\httpd.exe -k uninstall -n $apacheSvcName", $out, $err, $dir);
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Cannot install Apache service. Error:' . $stderr);
            }
            else if ($res == 0)
            {
                $stdout = null;
                Util::ExecC("sc config $apacheSvcName start= demand", $stdout );
                break;
            }
        }
        
        if ($i == $TRY_NUM)
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Cannot install Apache service');
        }

        //setup mysql
        $dir = realpath(SetupVars::get()->INSTALL_DIR . '/mysql/bin');
        $mysqlSvcName = '';
        for ($i = 0; $i < $TRY_NUM; $i ++)
        {
            $mysqlSvcName = "ad_mysql$i";
            $stdout = $stderr = '';
            $inifile = realpath(SetupVars::get()->INSTALL_DIR . '/mysql/my.cnf');
            $res = Util::Exec(".\\mysqld.exe --install $mysqlSvcName --defaults-file=\"$inifile\"", $stdout, $stderr, $dir);
            if (stripos($stdout, "denied") !== false)
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Access to SCM is denied');
            }
            else if (stripos($stdout, "successfully installed") !== false && $res == 0)
            {
                $stdout = null;
                Util::ExecC("sc config $mysqlSvcName start= demand", $stdout);
                break;
            }
        }

        if ($i == $TRY_NUM)
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Cannot install MySQL service');
        }

        $rep = array('@@APACHE_WIN_SVC_NAME@@' => $apacheSvcName , '@@MYSQL_WIN_SVC_NAME@@' => $mysqlSvcName);
        self::SubstituteFile(SetupVars::get()->INSTALL_DIR . '/AcquiaDrupalControlPanel/static.ini', $rep);
        self::$winSvcNames = array($apacheSvcName , $mysqlSvcName);
        
        return self::$winSvcNames;
    }
    
    private static function SetupXMail()
    {
        $sv = SetupVars::get();
        if( file_exists( $sv->INSTALL_DIR.'\xmail' ) )
        {
            $stdout = null;
            Util::ExecC( '"'.$sv->INSTALL_DIR.'\xmail\XMail.exe" --install-auto', $stdout );
            Util::ExecC( 'reg add HKLM\Software\GNU\XMail /f /v MAIL_ROOT /t REG_MULTI_SZ /d "'.$sv->INSTALL_DIR.'\xmail\MailRoot"', $stdout );
            Util::ExecC( 'reg add HKLM\Software\GNU\XMail /f /v MAIL_CMD_LINE /t REG_MULTI_SZ /d "-P- -B- -X- -Y- -F- -Mx 3"', $stdout);
            Util::ExecC( 'net start XMail', $stdout );
        }
    }

    /**
     * 
     *     
     * @param string $installDir
     * @return int - installation ID ,i.e. 0 for first stact installer on this machine, 1 for the second etc 
     */
    private static function SetupOSXServices( $installDir, $siteDir )
    {    
        $sv = SetupVars::get();
        
        $PLIST_DIR = self::PLIST_DIR;
        
        if( !file_exists($PLIST_DIR) )
        {
            if( @mkdir($PLIST_DIR, 0755 ) == false )
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . " Cannot create '$PLIST_DIR'");
        }
        
        //
        // Create users
        //
            
        $osGroups = OSXUtil::GetOSGroupList();
        if( !in_array( self::ACQUIA_GROUP, $osGroups ) )
            OSXUtil::CreateOSUserGroup( self::ACQUIA_GROUP );

        $osUsers = OSXUtil::GetOSUserList();
        if( !in_array( self::APACHE_USER, $osUsers ) )
            OSXUtil::CreateOSUser( self::APACHE_USER, '', self::ACQUIA_GROUP );
        if( !in_array( self::MYSQL_USER, $osUsers ) )
            OSXUtil::CreateOSUser( self::MYSQL_USER, '', self::ACQUIA_GROUP );           

        //
        // Register services
        //            
            
        $TRY_NUM = 20;
            
        $installNum = 0;
        for ( ; $installNum < $TRY_NUM; $installNum++ )
        {
            $plistFile = $PLIST_DIR.'/'.self::APACHE_PLIST.$installNum.'.plist';
            _Log("check if exists - '$plistFile'");
            if( !file_exists($plistFile) )
                break;   
        }
        if ($installNum == $TRY_NUM)
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Cannot create users. Try limit reached.');
        }

        //TODO: This is a temporary hack; after we integrate a Control Panel for OSX this information will be stored in Panel config file
        file_put_contents( $sv->INSTALL_DIR.'/common/instid', $installNum );             

        //
        // Setup MySQL
        //
        
        $stdout = null;
        
        //set mysql data directory permissions
        $cmdLine = sprintf("chown -R %s:%s ", self::MYSQL_USER, self::ACQUIA_GROUP)
           .escapeshellarg($installDir."/mysql/data");         
        Util::ExecC($cmdLine, $stdout);
        
        $svcName = self::MYSQL_PLIST.$installNum;
        $plistContents = str_replace(  
            array('@@SVC_NAME@@', '@@INSTALL_DIR@@', '@@USER@@'), 
            array($svcName, $installDir, self::MYSQL_USER), 
            $GLOBALS['DATA_OSX_MYSQL_PLIST'] ); 
    
        $plistFile = "$PLIST_DIR/$svcName.plist";
        if( !file_put_contents($plistFile, $plistContents) )
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Cannot create plist file'); 
    
        //install service
        $cmdLine = "launchctl load -w ".escapeshellarg($plistFile);
        Util::ExecC($cmdLine, $stdout);
        
        //
        // Setup Apache
        //
            
        //set sites directory access
        $cmdLine = "chmod a+w ".escapeshellarg($siteDir."/acquia-drupal/sites");         
        Util::ExecC($cmdLine, $stdout);
        // set logs folder access
        $cmdLine = sprintf("chown -R %s:%s ", self::APACHE_USER, self::ACQUIA_GROUP)
           .escapeshellarg($installDir."/apache/logs");         
        Util::ExecC($cmdLine, $stdout);
        
        
        //create plist file
        $svcName = self::APACHE_PLIST.$installNum;
        $plistContents = str_replace(  
            array('@@SVC_NAME@@', '@@INSTALL_DIR@@', '@@USER@@'), 
            array($svcName, $installDir, self::APACHE_USER), 
            $GLOBALS['DATA_OSX_APACHE_PLIST'] ); 
    
        $plistFile = "$PLIST_DIR/$svcName.plist";
        if( !file_put_contents($plistFile, $plistContents) )
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . ' Cannot create plist file');             
    
        //install service
        $cmdLine = "launchctl load -w ".escapeshellarg($plistFile);
        Util::ExecC($cmdLine, $stdout);     
    
        return $installNum;
    }

    
    private static function SetupUnixMysql ()
    {
/*      commented out as we ship precreated 'mysql' db with the stack
        $bindir = SetupVars::get()->INSTALL_DIR . '/mysql/bin';
        rename($bindir . '/mysql_install_db.sh', $bindir . '/mysql_install_db');
        chmod($bindir . '/mysql_install_db', 0755);
        $pairs = array(
                '@prefix@' => SetupVars::get()->INSTALL_DIR . '/mysql',
                '@bindir@' => SetupVars::get()->INSTALL_DIR . '/mysql/bin' ,
                '@libexecdir@' => SetupVars::get()->INSTALL_DIR . '/mysql/bin' ,
                '@localstatedir@' => SetupVars::get()->INSTALL_DIR . '/mysql/data',
                '@pkgdatadir@' => SetupVars::get()->INSTALL_DIR . '/mysql/share/mysql', 
                '@HOSTNAME@' => 'hostname');
        self::SubstituteFile($bindir . '/mysql_install_db', $pairs);
        $cmd = $bindir.'/mysql_install_db';
        $stdout = '';
        Util::ExecC($cmd, $stdout);
*/
    }

    private static function SetupUnixApache ()
    {
        $bindir = SetupVars::get()->INSTALL_DIR . '/apache/bin';
        //rename($bindir . '/apachectl.in', $bindir . '/apachectl');
        chmod($bindir . '/apachectl', 0755);
/*
        $pairs = array(
            '@exp_sbindir@' => SetupVars::get()->INSTALL_DIR . '/apache/bin' , 
            '@progname@' => 'httpd -f ' . SetupVars::get()->INSTALL_DIR . '/apache/conf/httpd.conf' , 
            '@LYNX_PATH@' => 'lynx' , 
            '@PORT@' => SetupVars::get()->APACHE_PORT , 
            '@APACHECTL_ULIMIT@' => '');
        self::SubstituteFile($bindir . '/apachectl', $pairs);
*/
        self::SubstituteFile( $bindir . '/apachectl' );
    }
    
    
    private static function SetupUnixPhpize()
    {
        $sv = SetupVars::get();
        // find php version
        $phpBin = $sv->INSTALL_DIR.'/php/bin/php';        
        
        $stderr = $stdout = null;
        Util::Exec( "$phpBin -v", $stdout, $stderr );
        
        $matches = null;                
        preg_match( '/^PHP\s+([\d\.]+)/', $stdout, $matches );
        $phpVer = $matches[1];
        
        // find php version id        
        $phpVerId = vsprintf( '%d%02d%02d', explode( '.', $phpVer ) );        

        $instDir = $sv->INSTALL_DIR.'/common';
        // replace vars in php-config
        $repl = array(
            '@prefix@'                => $sv->INSTALL_DIR.'/php',
            '@SED@'                   => '/usr/bin/sed',            
            '@exec_prefix@'           => '${prefix}',
            '@PHP_VERSION@'           => $phpVer,
            '@PHP_VERSION_ID@'        => $phpVerId,
            '@includedir@'            => '${prefix}/include', 
            '@PHP_LDFLAGS@'           => " -L {$sv->INSTALL_DIR}/common/lib",
            '@EXTRA_LIBS@'            => ' -lz -lmysqlclient -liconv -liconv -lpng -lz -ljpeg -lssl -lcrypto -lcurl -lbz2 -lz -lssl -lcrypto -lm  -lxml2 -lz -licucore -lm -lcurl -lz -lxml2 -lz -licucore -lm -lmysqlclient -lz -lm -lmysqlclient -lz -lm -lxml2 -lz -licucore -lm -lxml2 -lz -licucore -lm -lxml2 -lz -licucore -lm -lxml2 -lz -licucore -lm ',
            '@EXTENSION_DIR@'         => "{$sv->INSTALL_DIR}/php/ext",
            '@program_prefix@'        => '',
            '@program_suffix@'        => '',
            '@EXEEXT@'                => '',
            '@CONFIGURE_OPTIONS@'     => " '--prefix={$sv->INSTALL_DIR}/php' '--with-gd' '--with-png-dir=$instDir' '--with-curl=$instDir' '--enable-mbstring' '--enable-pcntl' '--enable-ftp' '--with-zlib' '--with-bz2' '--enable-zip' '--with-openssl=$instDir' '--with-mysql=$instDir'  '--with-pdo-mysql=$instDir'",
            '@PHP_INSTALLED_SAPIS@'   => "",
            '@bindir@'                => '${exec_prefix}/bin');       
        _Log("SetupUnixPhpize(): replace in php-config\n".print_r($repl, true) );
        $target = $sv->INSTALL_DIR.'/php/bin/php-config';
        self::SubstituteFile( $target.'.in', $repl );
        rename( $target.'.in', $target );
        chmod($target, 0777 );
        
        $repl = array(
            '@prefix@'      => $sv->INSTALL_DIR.'/php',
            '@exec_prefix@' => '${prefix}',
            '@libdir@'      => '${exec_prefix}/lib/php',
            '@includedir@'  => '${prefix}/include',
            '@SED@'         => '/usr/bin/sed' );
        _Log("SetupUnixPhpize(): replace in phpize\n".print_r($repl, true) );
        $target = $sv->INSTALL_DIR.'/php/bin/phpize';
        self::SubstituteFile( $target.'.in', $repl );
        rename( $target.'.in', $target );
        chmod($target, 0777 );
    }
    
    
    private static function SubstituteFile ($filename, $pairs = array())
    {
        _Log('Making substitutions in "%s"', $filename);
        if (empty($pairs))
        {
            $sv = SetupVars::get();
            $rc = new ReflectionClass('SetupVars');
            $props = $rc->getProperties();
            foreach ($props as $prop)
            {
                /*@var $prop ReflectionPropery*/
                if ($prop->isPublic() && ! $prop->isStatic())
                    $pairs['@@' . $prop->getName() . '@@'] = $prop->getValue($sv);
            }
        }
        $realname = realpath($filename);
        $data = file_get_contents($realname);
        if ($data === false)
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Cannot read ' . $filename);
        $data = str_replace(array_keys($pairs), array_values($pairs), $data);
        if (file_put_contents($realname, $data) === false)
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Cannot write ' . $filename);
    }

    private static function UnixPath ($path)
    {
        if (SetupVars::get()->PLATFORM_NAME == 'windows')
            $path = str_replace('\\', '/', $path);
        else if (SetupVars::get()->PLATFORM_NAME == 'osx')
            $path = str_replace(':', '/', $path);
        return rtrim($path, '/');
    }

    private static function StartMySQL ()
    {
        _Log("Starting Mysql");
        $cmdline = '';
        if (SetupVars::get()->PLATFORM_NAME == 'windows')
        {
            $cmdline = 'net start '.self::$winSvcNames[1];
            $stdout = $strerr = '';
            $r = Util::Exec($cmdline, $stdout, $strerr);
            if ( $r != 0 )
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unable to start MySQL. Stderr:' . $strerr);
            }
            self::WaitMySQLStart();
            /*
            $cmdline = sprintf ( 'start "mysql" "%s" --defaults-file="%s" --console', 
                realpath ( SetupVars::get()->INSTALL_DIR . '/mysql/bin/mysqld.exe' ), 
                realpath ( SetupVars::get()->INSTALL_DIR . '/mysql/my.cnf' ) );
*/
        }
        else if (in_array(SetupVars::get()->PLATFORM_NAME, array('osx' , 'linux')))
        {
            $rootopt = (trim(`whoami`) == 'root' ? '--user=root' : '');
            $cmdline = sprintf('"%s" --defaults-file="%s" %s', realpath(SetupVars::get()->INSTALL_DIR . '/mysql/bin/mysqld'), realpath(SetupVars::get()->INSTALL_DIR . '/mysql/my.cnf'), $rootopt);
            //
            // Start MySQL process 
            //
            $descriptorspec = array(0 => array("pipe" , "r") , // stdin is a pipe that the child will read from
                1 => array("pipe" , "w") , // stdout is a pipe that the child will write to
                2 => array("pipe" , "w")); // stderr is a file to write to            
            self::$hMysqlProc = proc_open($cmdline, $descriptorspec, $pipes);
            
            if (self::$hMysqlProc === false)
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Error executing start command');
            }
            else
            {
                fclose($pipes[0]);
                fclose($pipes[1]);
                fclose($pipes[2]);
            }
            self::WaitMySQLStart();
        }
        else
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unsupported platform:' . SetupVars::get()->PLATFORM_NAME);
        }
    }

    private static function StartApache ()
    {
        _Log("Starting Apache");
        $cmdline = '';
        if (SetupVars::get()->PLATFORM_NAME == 'windows')
        {
            $cmdline = 'net start ' . self::$winSvcNames[0];
            $stdout = $strerr = '';
            $r = Util::Exec($cmdline, $stdout, $strerr);
            if ( $r != 0 )
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unable to start Apache. Stderr:' . $strerr);
            }
        }
        else
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . 'Unsupported platform');
        }
    }

    private static function WaitMySQLStart ()
    {
        //
        // try to connect
        //
        _Log("Waiting for mysql to start");
        $oldTimeout = ini_get('mysql.connect_timeout');
        $hDB = null;
        ini_set('mysql.connect_timeout', 2);
        for ($i = 0; $i < 30; $i ++)
        {
            $hDB = @mysql_connect('127.0.0.1:' . SetupVars::get()->MYSQL_PORT, 'root', '');
            if ($hDB !== false)
                break;
            if (in_array(SetupVars::get()->PLATFORM_NAME, array('osx' , 'linux')))
            {
                sleep(2);
            }
        }
        ini_set('mysql.connect_timeout', $oldTimeout);
        if ($hDB == false)
        {
            _Log("Wait failed");
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unable to start MySQL');
        }
        else
        {
            _Log("Wait succeeded");
            mysql_close($hDB);
        }
    }    

    private static function StopMySQL ()
    {
        _Log("Stop MySQL");
        if (SetupVars::get()->PLATFORM_NAME == 'windows')
        {
            $cmdline = 'net stop ' . self::$winSvcNames[1];
            $stdout = $strerr = '';
            $r = Util::Exec($cmdline, $stdout, $strerr);
            if (/*strpos($stdout, 'stopped successfully') === false*/ $r != 0)
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unable to stop MySQL. Stderr:' . $strerr);
            }
        }
        else if (in_array(SetupVars::get()->PLATFORM_NAME, array('osx' , 'linux')))
        {
            $cmdLine = '';
            $params = ' --defaults-file="' . realpath(SetupVars::get()->INSTALL_DIR . '/mysql/my.cnf') . '" shutdown';
            $dir = realpath(SetupVars::get()->INSTALL_DIR . '/mysql/bin');
            $cmdLine = './mysqladmin' . ' ' . $params;
            $stdout = '';
            Util::ExecC($cmdLine, $stdout, $dir);
        }
        else
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unsupported platform:' . SetupVars::get()->PLATFORM_NAME);
        }
    }

    private static function WaitApacheStart ()
    {
        _Log('Wait for apache to start');
        $sv = SetupVars::get();        
        $timeout = 30;
        $startTime = time();
        for(;;)
        {
            $errno = 0;
            $errstr = '';
            $fs = @fsockopen( 'localhost', $sv->APACHE_PORT, $errno, $errstr, 1 );
            if( $fs !== FALSE )
            {
                _Log("Wait succeeded");
                fclose( $fs );
                break;
            }
            if( time() - $startTime > $timeout )
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Apache start timeout');
            sleep( 1 );
        }
    }
    
    private static function StopApache ()
    {
        _Log("Stop Apache");
        if (SetupVars::get()->PLATFORM_NAME == 'windows')
        {
            $cmdline = 'net stop ' . self::$winSvcNames[0];
            $stdout = $strerr = '';
            $r = Util::Exec($cmdline, $stdout, $strerr);
            if (/*strpos($stdout, 'stopped successfully') === false*/ $r != 0 )
            {
                throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unable to stop Apache. Stderr:' . $strerr);
            }
        }
        else
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . 'Unsupported platform');
        }
    }

    private static function InstallDrupal ()
    {
        _Log("Running Drupal web installer...");        
        $args = array(
            'profile' => 'acquia',
            'db_host' => '127.0.0.1', 
            'db_path' => SetupVars::get()->DB_NAME , 
            'db_user' => SetupVars::get()->DB_USERNAME , 
            'db_pass' => SetupVars::get()->DB_PASSWORD ,
            'db_port' => SetupVars::get()->MYSQL_PORT , 
            'site_name' => SetupVars::get()->SITE_NAME , 
            'username' => SetupVars::get()->ADM_USERNAME , 
            'password' => SetupVars::get()->ADM_PASSWORD , 
            'email' => SetupVars::get()->ADM_EMAIL , 
            'baseurl' => 'http://localhost:' . SetupVars::get()->APACHE_PORT.'/' /*. '/acquia-drupal/'*/ );
        install_drupal($args);
    }

    private static function CreateDB ()
    {
        global $DB_DUMP;
        //
        // connect
        //
        $hDB = @mysql_connect('127.0.0.1:' . SetupVars::get()->MYSQL_PORT, 'root', '');
        if ($hDB === false)
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . '() Unable to connect to MySQL');
        //
        // create database
        //
        $dbName = mysql_real_escape_string(SetupVars::get()->DB_NAME, $hDB);
        $dbUser = mysql_real_escape_string(SetupVars::get()->DB_USERNAME, $hDB);
        $dbPass = mysql_real_escape_string(SetupVars::get()->DB_PASSWORD, $hDB);
        self::MySQLQuery("CREATE DATABASE `$dbName`", $hDB);
        self::MySQLQuery("GRANT ALL PRIVILEGES ON *.* TO `$dbUser`@'localhost' IDENTIFIED BY '$dbPass'", $hDB);
        self::MySQLQuery("GRANT GRANT OPTION ON *.* TO `$dbUser`@'localhost'", $hDB);        
        self::MySQLQuery("FLUSH PRIVILEGES", $hDB);
        //
        // disconnect
        //
        mysql_close($hDB);
    }

    private static function MySQLQuery ($query, $hDB)
    {
        _Log("Runing SQL query:" . $query);
        $res = mysql_query($query, $hDB);
        if ($res === false)
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . "() Query failed.\nQuery:" . substr($query, 0, 200) . "\nError:" . mysql_error($hDB));
        }
        return $res;
    }
    private static $hMysqlProc;
    private static $winSvcNames;
}


class Util
{

    public static function IsWin ()
    {
        return (strtoupper(substr(PHP_OS, 0, 3)) === 'WIN');
    }
    public static function IsOSX ()
    {
        return (strtoupper(substr(PHP_OS, 0, 6)) === 'DARWIN');
    }
    
    public static function Exec ($cmd, &$stdout, &$stderr, $cwd = null)
    {
        $stdout = $stderr = null;
        _Log("Exec:" . $cmd);
        $tmpName = tempnam(sys_get_temp_dir(), 'drp');
        $descriptorspec = array(0 => array("pipe" , "r") , //stdin  
1 => array("pipe" , "w") , //stdout
2 => array("file" , $tmpName , "w")); //stderr
        $pipes = array();
        $process = proc_open($cmd, $descriptorspec, $pipes, $cwd);
        if (is_resource($process))
        {
            fclose($pipes[0]);
            while (! feof($pipes[1]))
            {
                $stdout .= stream_get_contents($pipes[1], 1024);
            }
            fclose($pipes[1]);
            $res = proc_close($process);
        }
        else
        {
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . sprintf(' Unable to execute command line - %s', $cmd));
        }
        $stderr = file_get_contents($tmpName);
        @unlink($tmpName);
        _Log("Exec result:" . $res);
        _Log("Exec stdout:" . $stdout);
        _Log("Exec stderr:" . $stderr);
        return $res;
    }

    public static function ExecC ($cmd, &$stdout, $cwd = null)
    {
        $stderr = '';
        $r = self::Exec($cmd, $stdout, $stderr, $cwd);
        if ($r != 0)
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . "Command returned " . $r . ". stderr:" . $stderr);
    }

    /**
     * Parses ini file
     * Return value format: $res[section_name][var_name]=var_value  
     * 
     * @param string $file
     * @return array
     */
    public static function ParseIniFile ($file)
    {
        $lines = @file($file);
        if ($lines === false)
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . "Cannot read:" . $file);
        $res = array();
        $curSection = null;
        foreach ($lines as $line)
        {
            $line = trim($line);
            if (! empty($line))
            {
                $matches = null;
                if (preg_match('/^\[(.*)\]$/', $line, $matches) > 0)
                {
                    $res[$matches[1]] = array();
                    $curSection = &$res[$matches[1]];
                }
                elseif (preg_match('/^([^=]+)=(.*)$/', $line, $matches) > 0)
                {
                    $curSection[$matches[1]] = $matches[2];
                }
            }
        }
        return $res;
    }
    
    public static function RmDir($dir)
    {
        $res = false;        
        if( is_dir($dir)  )
        {
            if( self::IsWin() )
            {
                $res = ( self::Exec( "rmdir /S /Q ".escapeshellarg($dir), $stdout, $stderr ) == 0 );    
            }
            else if( self::IsOSX() )
            {
                $res = ( self::Exec( "rm -r ".escapeshellarg($dir), $stdout, $stderr ) == 0 );
            }
            else
            {
                die("unknown os");
            }            
        }
        return $res;
    }    

    public static function ParseArgv( $argv )
    {
        $res = array();
        foreach ( $argv as $a )
        {
            list( $nm, $val ) = explode( '=', $a );
            if( substr($nm, 0, 2) == '--' && !empty($val) )
            {
                $nm = trim( $nm, '-' );
                $res[$nm] = $val;                
            }
        }
        return $res;
    }
    
    public static function RegexReplaceInFile( $file, $regex, $repl )
    {
        _Log("RegexReplaceInFile. file:$file, regex:$regex, repl:$repl" );
        $text = file_get_contents( $file );
        if( $text === false )
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . sprintf(' Unable read file - %s', $file) );
        $count;
        $text = preg_replace( $regex, $repl, $text, -1, $count );
        _Log("Nummatch:$count");        
        if( $count > 0  )
        {
//            _Log("text: %s", $text );
            file_put_contents( $file, $text );
        }
    }
}



class OSXUtil
{    
    
    public static function GetOSUserList()
    {
        _Log(__CLASS__.'::'.__FUNCTION__);
        $stdout = '';
        Util::ExecC( "dscl localhost -list /Local/Default/Users", $stdout );
        return explode( "\n", $stdout );    
    }
        
    
    public static function GetOSGroupList()
    {
        _Log(__CLASS__.'::'.__FUNCTION__);
        $stdout = '';
        Util::ExecC( "dscl localhost -list /Local/Default/Groups", $stdout );
        return explode( "\n", $stdout );    
    }
    
    public static function GetOSNewGroupPrimaryID()
    {
        $stdout = '';
        Util::ExecC( "dscl localhost -list /Local/Default/Groups PrimaryGroupID | awk '{print $2}'", $stdout );
        $ids = explode( "\n", $stdout );
        sort( $ids, SORT_NUMERIC );
        $unqueId = $ids[count($ids) - 1] + 1;
        return $unqueId;            
    }
    
    public static function GetOSNewUserUniqueID()
    {
        $stdout = '';
        Util::ExecC( "dscl localhost -list /Local/Default/Users UniqueID | awk '{print $2}'", $stdout );
        $ids = explode( "\n", $stdout );
        sort( $ids, SORT_NUMERIC );
        $unqueId = $ids[count($ids) - 1] + 1;
        return $unqueId;
    }
    
    /**
     * @return created group PrimaryGroupID
     */
    public static function CreateOSUserGroup( $groupName )
    {
        _Log(__CLASS__.'::'.__FUNCTION__."($groupName)");
        if( self::OSGroupExists($groupName) )
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . " group already exists - ".$groupName );                
        
        $stdout = '';
        $gid = self::GetOSNewGroupPrimaryID();
        Util::ExecC( "dscl localhost -create /Local/Default/Groups/$groupName", $stdout );
        Util::ExecC( "dscl localhost -create /Local/Default/Groups/$groupName PrimaryGroupID $gid", $stdout );
        return intval( $gid );        
    }
    
    /**
     * @param string $userName
     * @return int - group PrimaryGroupID on success and null on failure
     */
    public static function GetOSUserGroupID( $userName )
    {
        $gid = self::GetDSPropertyVal("/Local/Default/Users/$userName", 'PrimaryGroupID');
        return $gid === null? null: intval($gid);        
    }
    
    /**
     * @param int $groupID
     * @return string - group name on success and null otherwise
     */
    public static function GetOSGroupNameByID( $groupID )
    {        
        $gnames = self::GetDSNodeListByPropertyKeyVal( '/Local/Default/Groups', 'PrimaryGroupID', $groupID );
        return empty( $gnames )? null: $gnames[0];   
    }
        
    public static function DeleteOSUserGroup( $groupName )
    {
        if( !self::OSGroupExists($groupName) )
            throw new Exception(__CLASS__ . '::' . __FUNCTION__ . " Group doesn't exist - ".$groupName );                
        
        $stdout = null;
        Util::ExecC( "dscl localhost -delete /Local/Default/Groups/$groupName", $stdout );              
    }
    
    /**
     * @param string $userName
     * @param string $password
     * @param string $group
     * @return created user ID
     */
    public static function CreateOSUser( $userName, $password, $group )
    {
        _Log("Ceate user(name:$userName password:$password, group:$group )");
        $uid = -1;
        Util::ExecC( "dscl localhost -create /Local/Default/Users/$userName", $stdout );
        $uid = self::GetOSNewUserUniqueID();
        Util::ExecC( "dscl localhost -create /Local/Default/Users/$userName UniqueID $uid", $stdout );  
        $gid = self::GetDSPropertyVal("/Local/Default/Groups/$group", 'PrimaryGroupID');
        Util::ExecC( "dscl localhost -create /Local/Default/Users/$userName PrimaryGroupID $gid", $stdout );
        Util::ExecC( "dscl localhost -create /Local/Default/Users/$userName UserShell /usr/bin/false", $stdout );
        Util::ExecC( "dscl localhost -passwd /Local/Default/Users/$userName $password", $stdout );
        
        
        self::GroupMembershipUserAdd( $group, $userName );
        
        return $uid;        
    }

    public static function DeleteOSUser( $userName )
    {
        $stdout = '';
        
        if( !self::OSUserExists( $userName ) )
            throw new Exception(__CLASS__.'::'.__FUNCTION__ . " User doesn't exist - ".$userName );
        
        
        // delete the user from primary group        
        $gid = self::GetDSPropertyVal( "/Local/Default/Users/$userName", 'PrimaryGroupID' );
        if( $gid !== null  )
        {
            
            $gname = self::GetOSGroupNameByID( $gid );
            if( $gname )
            {
                self::GroupMembershipUserRemove( $gname, $userName );
            }
        }

        Util::ExecC( "dscl localhost -delete /Local/Default/Users/$userName", $stdout );              
    }

    private static function GetGroupMembership( $groupName )
    {
        $res = null;
        $users = self::GetDSPropertyVal("/Local/Default/Groups/$groupName", 'GroupMembership'); 
        return empty($users)? array(): explode( ' ', $users );
    }

    private static function GroupMembershipUserAdd( $group, $user )
    {
        $groupUsers = self::GetGroupMembership( $group );
        if( !in_array( $user, $groupUsers ) )
            Util::ExecC( "dscl localhost -append /Local/Default/Groups/$group GroupMembership $user", $stdout );
    }
    
    public static function GroupMembershipUserRemove( $group, $user )
    {
        $groupUsers = self::GetGroupMembership( $group );
        if( in_array( $user, $groupUsers ) )
            Util::ExecC( "dscl localhost -delete /Local/Default/Groups/$group GroupMembership $user", $stdout );
    }
    
    private static function PathExists( $path )
    {
        $stdout = $stderr = '';
        $r = Util::Exec( "dscl localhost -read $path $key", $stdout, $stderr );
        return $r == 0;
    }
    
    public static function OSUserExists( $user )
    {
        return self::PathExists( "/Local/Default/Users/$user" );
    }

    public static function OSGroupExists( $user )
    {
        return self::PathExists( "/Local/Default/Groups/$user" );
    }
    
    
    private static function GetDSPropertyVal($path, $key)
    {
        $stdout = '';
        Util::ExecC( "dscl localhost -read $path $key", $stdout );        
        $matches = null;
        $result = null;        
        if( preg_match( "/^$key: ?(.*)$/s", $stdout, $matches ) > 0 )
        {
            $result = trim($matches[1]);
        }
        return $result;
    }
    
    private static function GetDSNodeListByPropertyKeyVal( $path, $key, $val )
    {
        $stdout = '';
        Util::ExecC( "dscl localhost -list $path $key", $stdout );
        
        $res = array();
        $matches = null;
        $r = preg_match_all( '/^(\S+)\s+(\S*)$/m', $stdout, $m );
        
        for( $i = 0; $i < $r; $i++ )
            if( $m[2][$i] == $val )
                $res[] = $m[1][$i];
        return $res;
    }
}


class Uninst
{

    public static function PreUninst ()
    {
        if (Util::IsWin())
        {
            self::UnregisterWindowsServices();
        }
        if( Util::IsOSX() )
        {
            self::UnregisterOSXServices();
        }
        self::RemoveHostEntries();
    }

    
    public  static function UnregisterWindowsServices ()
    {
        $installDir = self::InstallDir();
        
        $ini = Util::ParseIniFile($installDir . '/AcquiaDrupalControlPanel/static.ini');
        $stdout = $stderr = '';
                
        if( $ini['apache']['winServiceName'][0] != '@' )
        {
            Util::Exec('net stop ' . $ini['apache']['winServiceName'], $stdout, $stderr);
            $cmd = escapeshellarg($installDir."\\".$ini['apache']['executablePath']) . ' -k uninstall -n ' . $ini['apache']['winServiceName'];
            Util::Exec($cmd, $stdout, $stderr);
        }
        
        if( $ini['mysql']['winServiceName'] )
        {
            Util::Exec('net stop ' . $ini['mysql']['winServiceName'], $stdout, $stderr);
            $cmd = escapeshellarg($installDir."\\".$ini['mysql']['executablePath']) . ' --remove ' . $ini['mysql']['winServiceName'];
            Util::Exec($cmd, $stdout, $stderr);            
        }

        if( file_exists($installDir.'\xmail') )
        {
            Util::Exec('net stop XMail', $stdout, $stderr);
            Util::Exec( '"'.$installDir.'\xmail\XMail.exe" --remove');
        }
    }

    
    public static function UnregisterOSXServices ()
    {
        $installDir = self::InstallDir();
        
        $instId = file_get_contents($installDir.'/common/instid');
        if( $instId !== FALSE )
        {
            @unlink($installDir.'/common/instid');
            $stdout = $stderr = '';
            
            $apachePlist = Setup::PLIST_DIR.'/'.Setup::APACHE_PLIST.$instId.'.plist';
            $mysqlPlist = Setup::PLIST_DIR.'/'.Setup::MYSQL_PLIST.$instId.'.plist';
            
            Util::Exec("launchctl unload -w $apachePlist", $stdout, $stderr );
            Util::Exec("launchctl unload -w $mysqlPlist", $stdout, $stderr );
            
            unlink( $apachePlist );
            unlink( $mysqlPlist );

            $lastInstall = true;
            if ($handle = opendir(Setup::PLIST_DIR)) 
            {
                while (false !== ($file = readdir($handle))) 
                {
                    if( strpos( $file, Setup::APACHE_PLIST ) === 0 || 
                        strpos( $file, Setup::MYSQL_PLIST ) === 0  )
                    {
                        $lastInstall = false;
                        break;                            
                    }
                }
                closedir($handle);
            }
            
            if( $lastInstall )            
            {
                try{ OSXUtil::DeleteOSUser( Setup::APACHE_USER ); } catch (Exception $e){}    
                try{ OSXUtil::DeleteOSUser( Setup::MYSQL_USER); } catch (Exception $e){}
                try{ OSXUtil::DeleteOSUserGroup( Setup::ACQUIA_GROUP ); } catch (Exception $e){}
            }                   
        }
    }
    
    
    function RemoveHostEntries()
    {
        $installDir = self::InstallDir();

        $hostsfile = null;        
        if( Util::IsWin() )
        {
            $hostsfile = getenv('SystemRoot')."\\system32\\drivers\\etc\\hosts";
            $iniFile = $installDir . '/AcquiaDrupalControlPanel/dynamic.ini';
        }
        else if( Util::IsOSX() )
        {
            $hostsfile = "/etc/hosts";
            $iniFile = $installDir . '/Acquia Drupal Control Panel.app/Contents/MacOS/dynamic.ini';
        }
        else
        {
            die('?');
        }
            
            
        $ini = Util::ParseIniFile($iniFile);
        $hosts = array();    
        foreach ( $ini as $section => $vars )
        {
            $secPath = explode( '/', $section );
            if( count($secPath) == 3 && $secPath[0] == 'sites' && $secPath[1] == 'm_sites' && $vars['host'] != 'localhost' )
            {
                $hosts[] = $vars['host']; 
            }
        }

        if( count($hosts) > 0 )
        {
            $lines = file( $hostsfile );
            $lines2 = array();
            foreach ( $lines as $line )
            {
                $remove = false;
                
                foreach ( $hosts as $host )
                {
                    if( preg_match('/^127.0.0.1\s+'.preg_quote($host).'/', $line) )
                    {
                        $remove = true;
                        break;
                    }
                }
                if( !$remove )
                    $lines2[] = $line;
            }
            file_put_contents( $hostsfile, implode( "", $lines2 ) );
            //echo implode( "", $lines2 );
        }
        
    }
    
    private static function InstallDir()
    {
        return realpath( dirname(__FILE__) . '/../..' );
    }
}

function main ($argv)
{
    try
    {
        if (isset($argv[1]) && $argv[1] == 'preuninst')
        {
            Uninst::PreUninst();
        }
        else if( isset($argv[1]) && $argv[1] == 'finishinst' )
        {
            array_shift($argv);
            array_shift($argv);

            Setup::FinishSiteInstall( Util::ParseArgv( $argv ) ); 
        }
        else
        {
            Setup::Run();
        }
    }
    catch (Exception $ex)
    {
        echo $ex->getMessage();
        _Log("Exception caught:" . $ex->getMessage());
        
        return -1;
    }
    echo SUCCESS_STRING;
    return 0;
}


//
// Drupal installer
//
/** Helper for CLI driven scripts.
 */
class DrupalCli
{

    /**
     * Read and setup _SERVER env for running drupal code from CLI
     * 
     * Example:
     *  DrupalCli::Setup(&$_SERVER)
     *
     * @param array $env $_SERVER
     */
    static public function Setup (&$env)
    {
        $host = 'localhost';
        $path = '';
        $args = DrupalCli::ParseArguments($env['argv']);
        if (isset($args['url']))
        {
            $parse = parse_url($args['url']);
            $host = $parse['host'];
            $path = (array_key_exists('path', $parse) ? $parse['path'] : '');
            if (isset($parse['port']) && is_numeric($parse['port']))
            {
                $env['SERVER_PORT'] = $parse['port'];
                $host .= ':' . $parse['port'];
            }
        }
        # useful for debugger where you cannot control current directory
        if (isset($args['chdir']))
        {
            chdir($args['chdir']);
        }
        $env['HTTP_HOST'] = $host;
        $env['REMOTE_ADDR'] = '127.0.0.1';
        $env['SERVER_ADDR'] = '127.0.0.1';
        $env['SERVER_SOFTWARE'] = 'PHP/curl';
        $env['SERVER_NAME'] = 'localhost';
        $env['REQUEST_URI'] = $path . '/';
        $env['REQUEST_METHOD'] = 'GET';
        $env['SCRIPT_NAME'] = $path . '/index.php';
        $env['PHP_SELF'] = $path . '/index.php';
        $env['HTTP_USER_AGENT'] = 'Drupal command line';
    }

    /**
     * Parses simple parameters from CLI. 
     * 
     * Puts trailing parameters into string array in 'extraArguments' 
     * 
     * Example:
     * $args = DrupalCli::ParseArguments($_SERVER['argv']);
     * if ($args['verbose']) echo "Verbose Mode On\n";
     * $files = $args['extraArguments'];
     *
     * Example CLI:
     *  --foo=blah -x -h  some trailing arguments 
     * 
     * if multiValueMode is true 
     * Example CLI:
     *  --include=a --include=b --exclude=c 
     * Then
     *  $args = DrupalCli::ParseArguments($_SERVER['argv']);
     *  $args['include[]'] will equal array('a', 'b')
     *  $args['exclude[]'] will equal array('c')
     *  $args['exclude'] will equal c
     *  $args['include'] will equal b   NOTE: only keeps last value
     * 
     * @param unknown_type $argv
     * @param supportMutliValue - will store 2nd copy of value in an array with key "foo[]" 
     * @return unknown
     */
    static public function ParseArguments ($argv, $mutliValueMode = False)
    {
        $args = array();
        $args['extraArguments'] = array();
        array_shift($argv); // scriptname
        foreach ($argv as $arg)
        {
            if (ereg('^--([^=]+)=(.*)', $arg, $reg))
            {
                $args[$reg[1]] = $reg[2];
                if ($mutliValueMode)
                {
                    DrupalCli::addItemAsArray($args, $reg[1], $reg[2]);
                }
            }
            elseif (ereg('^[-]{1,2}([^[:blank:]]+)', $arg, $reg))
            {
                $nonnull = '';
                $args[$reg[1]] = $nonnull;
                if ($mutliValueMode)
                {
                    DrupalCli::addItemAsArray($args, $reg[1], $nonnull);
                }
            }
            else
            {
                $args['extraArguments'][] = $arg;
            }
        }
        return $args;
    }

    /**
     * Adds a value as an array of one, or appends to an existing array elements 
     *
     * @param unknown_type $array
     * @param unknown_type $item
     */
    static function addItemAsArray (&$array, $key, $item)
    {
        $array_key = $key . '[]';
        if (array_key_exists($array_key, $array))
        {
            $array[$array_key][] = $item;
        }
        else
        {
            $array[$array_key] = array($item);
        }
    }

    static function AssertInDocRoot ()
    {
        if (! file_exists('includes/bootstrap.inc'))
        {
            throw new Exception("Must be run from web docroot. Current working directory is:\n" . getcwd());
        }
    }
}


/**
 * Installs Drupal by screen scraping installer pages.  Only tests on D6 and D7
 */
class DrupalInstaller
{
    public $additionalParams = array();
    public $site_name;
    public $email;
    public $username;
    public $password;
    public $profile;
    public $db_path;
    public $db_user;
    public $db_pass;
    public $db_host;
    public $db_port;
    public $db_prefix;

    public function install ($baseurl)
    {
        $userAgent = new DrupalUserAgent($baseurl);
        $this->run(STDOUT, $userAgent);
    }
    
    public function install2 ($baseurl)
    {
        $userAgent = new DrupalUserAgent($baseurl);
        $this->run2(STDOUT, $userAgent);
    }
    
    public function run ($stdout, DrupalUserAgent $userAgent)
    {
        $i = 1;
        $done = False;
        // hit landing page first
        $userAgent->get("install.php");
        $i = 0;
        while (! $done)
        {
            if ($i ++ > 10)
            {
                _Log("DWI:$html");
                throw new Exception("Maximum number of calls made. Site may not be setup correctly.");
            }
            // Implies english locale
            $html = $userAgent->get("install.php?profile=" . $this->profile . "&locale=en&id=1&op=do_nojs");
            _Log("DWI: Step $i...\n");

            if (preg_match('/Database configuration/', $html))
            {
                $dom = DrupalUserAgent::asDom($html);
                $this->configureDatabase($stdout, $dom, $userAgent);
            }
            else
            {
                $done = preg_match('/Remaining 0/', $html);
            }
        }
        _Log("DWI:  Step $i...");
        $dom = $userAgent->get("install.php?profile=" . $this->profile . "&locale=en&id=1&op=finished");
        _Log("DWI: configuring ...");
        $dom = $userAgent->getDom("install.php?profile=" . $this->profile . "&locale=en&id=1");
        $form = DrupalFormLoader::findById($dom, "install-configure-form");
        if ($form == NULL)
        {
            _Log("DWI:".$dom->asXML());
            throw new Exception("Could not find installation form in installation screen");
        }
        $form->setValue('site_name', $this->site_name);
        $form->setValue('site_mail', $this->email);
        $form->setValue('account[name]', $this->username);
        $form->setValue('account[mail]', $this->email);
        $form->setValue('account[pass][pass1]', $this->password);
        $form->setValue('account[pass][pass2]', $this->password);
        $form->setValue('clean_url', '1');
        $fields = array_merge($form->getParams(), $this->additionalParams);
        $html = $userAgent->post($form->getAction(), $fields);
        if (! preg_match('/complete/', $html))
        {
            _Log("DWI:$html");
            throw new Exception('Configuration not complete');
        }
        _Log("DWI: finished.");
    }

    public function run2 ($stdout, DrupalUserAgent $userAgent)
    {
        $i = 1;
        $done = False;
        // hit landing page first
        $userAgent->get("install.php");
        $i = 0;
        while (! $done)
        {
            if ($i ++ > 10)
            {
                _Log("DWI:$html");
                throw new Exception("Maximum number of calls made. Site may not be setup correctly.");
            }
            // Implies english locale
            $html = $userAgent->get("install.php?profile=" . $this->profile . "&locale=en&id=1&op=do_nojs");
            _Log("DWI: Step $i...\n");

            $done = preg_match('/Remaining 0/', $html);
        }
        _Log("DWI:  Step $i...");
        $dom = $userAgent->get("install.php?profile=" . $this->profile . "&locale=en&id=1&op=finished");
        _Log("DWI: configuring ...");
        $dom = $userAgent->getDom("install.php?profile=" . $this->profile . "&locale=en&id=1");
        $form = DrupalFormLoader::findById($dom, "install-configure-form");
        if ($form == NULL)
        {
            _Log("DWI:".$dom->asXML());
            throw new Exception("Could not find installation form in installation screen");
        }
        $form->setValue('site_name', $this->site_name);
        $form->setValue('site_mail', $this->email);
        $form->setValue('account[name]', $this->username);
        $form->setValue('account[mail]', $this->email);
        $form->setValue('account[pass][pass1]', $this->password);
        $form->setValue('account[pass][pass2]', $this->password);
        $form->setValue('clean_url', '1');
        $fields = array_merge($form->getParams(), $this->additionalParams);
        $html = $userAgent->post($form->getAction(), $fields);
        if (! preg_match('/complete/', $html))
        {
            _Log("DWI:$html");
            throw new Exception('Configuration not complete');
        }
        _Log("DWI: finished.");
    }
    
    function configureDatabase ($stdout, $dom, DrupalUserAgent $userAgent)
    {
        $formId = "install-settings-form";
        $form = DrupalFormLoader::findById($dom, $formId);
        if ($form == NULL)
        {
            _Log("DWI:".$dom->asXML());
            throw new Exception("Could not find installation '$formId' form in installation screen");
        }
        $form->setValue('db_path', $this->db_path);
        $form->setValue('db_user', $this->db_user);
        $form->setValue('db_pass', $this->db_pass);
        $form->setValue('db_host', $this->db_host);
        $form->setValue('db_port', $this->db_port);
        $form->setValue('db_prefix', $this->db_prefix);
        $userAgent->post($form->getAction(), $form->getParams());
    }

    static function array_remove (&$array, $item)
    {
        $value = null;
        if (array_key_exists($item, $array))
        {
            $value = $array[$item];
            unset($array[$item]);
        }
        return $value;
    }
}


/**
 * GET and POST form on drupal, preserving existing form data
 */
class DrupalUserAgent
{
    private $handle;
    private $base_url;

    public function __construct ($base_url = '')
    {
        $this->handle = curl_init();
        $this->base_url = $base_url;
    }

    public function __destruct()
    {
        $this->close();
    }
    
    public function close ()
    {
        if( $this->handle )
        {
            curl_close($this->handle);
            unset($this->handle);            
        }
    }

    static public function asDom ($content)
    {
        $html = DOMDocument::loadHTML($content);
        if (is_null($html))
        {
            throw new Exception("Could not parse HTML at url " . $url);
        }
        return simplexml_import_dom($html);
    }

    public function getDom ($url)
    {
        $content = $this->get($url);
        return self::asDom($content);
    }

    public function get ($url)
    {   
        $url = str_replace(' ', '%20', $url);     
        _Log("DWI: GET(%s)", $url);
        $this->curlOpen($url);
        curl_setopt($this->handle, CURLOPT_HTTPGET, True);
        return $this->curlExecute();
    }

    public function post ($url, $params)
    {
        $url = str_replace(' ', '%20', $url);
        _Log("DWI: POST(%s). params:%s", $url, print_r($params, true) );
        $this->curlOpen($url);
        curl_setopt($this->handle, CURLOPT_POST, count($params));
        curl_setopt($this->handle, CURLOPT_POSTFIELDS, DrupalUserAgent::paramString($params));
        return $this->curlExecute();
    }

    public static function paramString ($params)
    {
        $post = array();
        foreach ($params as $field => $value)
        {
            $post[] = urlencode($field) . '=' . urlencode($value);
        }
        return implode($post, '&');
    }

    public function curlOpen ($url)
    {        
        $options = array(
            CURLOPT_URL => $this->absoluteUrl($url) , //CURLOPT_VERBOSE => TRUE,
            CURLOPT_COOKIEJAR => NULL , // accepts session cookie, just not on disk
            CURLOPT_TIMEOUT => 300 , 
            CURLOPT_FOLLOWLOCATION => TRUE , 
            CURLOPT_RETURNTRANSFER => TRUE , 
            CURLOPT_USERAGENT => 'drupal-user-agent');
        curl_setopt_array($this->handle, $options);
    }

    public function setCookies ($cookies)
    {
        curl_setopt($this->handle, CURLOPT_COOKIE, DrupalUserAgent::paramString($params));
    }

    public function curlExecute ()
    {
        //curl_setopt( $this->handle, CURLOPT_VERBOSE, TRUE );
        //$fp = fopen("/Applications/acquia-drupal/curl.log", "a");
        //curl_setopt( $this->handle, CURLOPT_STDERR, $fp );
        
        $content = curl_exec($this->handle);
        //fclose( $fp );
        $err_id = curl_errno($this->handle);
        if ($err_id != 0)
        {
            throw new Exception( sprintf( "curl error: %s (%d)", curl_error($this->handle), curl_errno($this->handle) ) );
        }
        _Log("DWI: returned html:\n========================\n$content\n========================\n\n\n" );
        return $content;
    }

    public function absoluteUrl ($url)
    {
        if (empty($url))
        {
            return $this->base_url;
        }
        if ( empty($this->base_url) || 
             strpos($url, $this->base_url) === 0 || 
             strtolower(substr($url, 0, 7)) == 'http://' )
        {
            return $url;
        }
        return $this->base_url . $url;
    }
}


class DrupalFormLoader
{
    private $form;
    private $params = array();

    public function __construct ($form)
    {
        $this->form = $form;
    }

    public static function findById ($dom, $id)
    {
        $elems = $dom->elements->xpath("//form[@id='" . $id . "']");
        return empty($elems) ? NULL : new DrupalFormLoader($elems[0]);
    }

    public static function findBySubmitValue ($dom, $submit)
    {
        $elems = $dom->elements->xpath("//input[@type='submit' and @value='" . $submit . "']");
        if (! $elems)
        {
            return NULL;
        }
        foreach ($elems as $elem)
        {
            $form = $elem->xpath('ancestor::form');
            if (! empty($form))
            {
                return new DrupalFormLoader($form);
            }
        }
    }

    public function setValue ($name, $value)
    {
        $this->params[$name] = $value;
    }

    public function getAction ()
    {
        return (string) $this->form['action'];
    }

    public function getParams ()
    {
        // surely there is more types, only understand these at the moment
        $elements = $this->form->xpath("descendant::input|descendant::textarea");
        foreach ($elements as $element)
        {
            // SimpleXML objects need string casting all the time.
            $name = (string) $element['name'];
            // ignore preset parameters
            if (empty($name) || /*array_key_exists($name, $this->params)*/ isset($this->params[$name]))
            {
                continue;
            }
            switch ($element->getName())
            {
            case 'input':
                $type = $element['type'];
                switch ($type)
                {
                case 'radio':
                case 'checkbox':
                    if (! isset($element['checked']))
                    {
                        break;
                    }
                // Deliberate no break.
                default:
                    $value = isset($element['value']) ? (string) $element['value'] : '';              break;
                }
                break;
            case 'textarea':
                $value = (string) $element;
                break;
            case 'select':
                throw new Exception('select element not supported yet');
            }
            $this->params[$name] = $value;
        }
        return $this->params;
    }
}

function install_drupal ($args)
{
    DrupalCli::Setup($_SERVER);
    $installer = new DrupalInstaller();
    $installer->profile = @$args['profile'];
    $installer->db_path = @$args['db_path'];
    $installer->db_user = @$args['db_user'];
    $installer->db_pass = @$args['db_pass'];
    $installer->db_host = @$args['db_host'];
    $installer->db_port = @$args['db_port'];
    $installer->db_prefix = @$args['db_prefix'];
    $installer->site_name = @$args['site_name'];
    $installer->username = @$args['username'];
    $installer->password = @$args['password'];
    $installer->email = @$args['email'];
    //$installer->additionalParams = $args;
    $installer->install(@$args['baseurl']);
}

/**
 * This function finishes installation of a partially installed site.
 * Actually just it drives the last installation page with sitename/username/password
 * Used by control panel to finish the site installation 
 */
function finish_drupal_install ($args)
{
}



$DATA_OSX_MYSQL_PLIST = <<<MYDATA
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>KeepAlive</key>
    <true/>
    <key>Label</key>
    <string>@@SVC_NAME@@</string>
    <key>OnDemand</key>
    <false/>
    <key>ProgramArguments</key>
    <array>
        <string>@@INSTALL_DIR@@/mysql/bin/mysqld</string>
        <string>--defaults-file=@@INSTALL_DIR@@/mysql/my.cnf</string>
    </array>
    <key>RunAtLoad</key>
    <true/>
    <key>UserName</key>
    <string>@@USER@@</string>
    <key>WorkingDirectory</key>
    <string>@@INSTALL_DIR@@/mysql/bin</string>
</dict>
</plist>    
MYDATA;

$DATA_OSX_APACHE_PLIST = <<<MYDATA
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>Label</key>
    <string>@@SVC_NAME@@</string>
    <key>OnDemand</key>
    <false/>
    <key>ProgramArguments</key>
    <array>
        <string>@@INSTALL_DIR@@/apache/bin/httpd</string>
        <string>-d</string>
        <string>@@INSTALL_DIR@@/apache</string>
        <string>-D</string>
        <string>FOREGROUND</string>
    </array>
    <key>EnvironmentVariables</key>
    <dict>
        <key>DYLD_LIBRARY_PATH</key>  <string>@@INSTALL_DIR@@/common/lib</string>
    </dict>
    
    <key>RunAtLoad</key>
    <true/>   
    <key>SHAuthorizationRight</key>
    <string>system.preferences</string>    
    <key>UserName</key>
    <string>@@USER@@</string>    
</dict>
</plist>
MYDATA;

$DATA_PHPINFO_PHP_CODE = <<<MYDATA
<?php
if( \$_SERVER['REMOTE_ADDR'] == '127.0.0.1' || \$_SERVER['REMOTE_ADDR'] == '::1')
{
    phpinfo();
}
?>
MYDATA;

//
// Script entry point
//

$r = main($_SERVER['argv']);
exit( $r );

//$sv = SetupVars::get();
//$sv->INSTALL_DIR = '/prj/trunk/csb/_tmp'; 
//Setup::SetupUnixPhpize();

?>